'''LSST JupyterHub utility functions.
'''
from .utils import (rreplace, sanitize_dict, get_execution_namespace,
                    make_logger, str_bool, str_true, listify, list_duplicates)
all = [rreplace, sanitize_dict, get_execution_namespace, make_logger,
       str_bool, str_true, listify, list_duplicates]



'''LSST Authenticator to use JWT token present in request headers.
'''
from jwtauthenticator.jwtauthenticator import JSONWebTokenAuthenticator
from .lsstauth import LSSTAuthenticator
from .lsstjwtloginhandler import LSSTJWTLoginHandler
from ..utils import make_logger


class LSSTJWTAuthenticator(LSSTAuthenticator, JSONWebTokenAuthenticator):
    auth_refresh_age = 900
    header_name = "X-Portal-Authorization"
    header_is_authorization = True
    username_claim_field = 'uid'

    def __init__(self, *args, **kwargs):
        '''Add LSST Manager structure to hold LSST-specific logic.
        '''
        self.log = make_logger()
        self.log.debug("Creating LSSTJWTAuthenticator")
        super().__init__(*args, **kwargs)

    def get_handlers(self, app):
        '''Install custom handlers.
        '''
        return [
            (r'/login', LSSTJWTLoginHandler),
        ]

    def logout_url(self, base_url):
        '''Returns the logout URL for JWT.
        '''
        return self.lsst_mgr.config.jwt_logout_url



#!/usr/bin/env python3
import argparse
from .reaper import Reaper


def standalone():
    '''Run as a standalone command.
    '''
    args = parse_args()
    wilford_grimly = Reaper(host=args.repo,
                            owner=args.owner,
                            name=args.name,
                            keep_dailies=args.dailies,
                            keep_weeklies=args.weeklies,
                            keep_experimentals=args.experimentals,
                            port=args.port,
                            cachefile=args.cachefile,
                            debug=args.debug)

    wilford_grimly.more_cowbell()


def parse_args():
    '''Parse command-line arguments.
    '''
    desc = "Remove obsolete lab images"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("-d", "--debug", action="store_true",
                        help="enable debugging")
    parser.add_argument("-r", "--repo", "--repository",
                        help="Docker repository host")
    parser.add_argument("-o", "--owner", "--organization", "--org",
                        help="repository owner [lsstsqre]",
                        default="lsstsqre")
    parser.add_argument("-n", "--name",
                        help="repository name [sciplat-lab]",
                        default="sciplat-lab")
    parser.add_argument("-q", "--dailies", "--daily", "--quotidian", type=int,
                        help="# of daily builds to keep [15]",
                        default=15)
    parser.add_argument("-w", "--weeklies", "--weekly", type=int,
                        help="# of weekly builds to keep [78]",
                        default=78)
    parser.add_argument("-e", "--experimentals", "--experimental",
                        "--exp", type=int,
                        help="# of experimental builds to keep [10]",
                        default=10)
    parser.add_argument("-p", "--port", help="Repository port [443 for" +
                        " secure, 80 for insecure]",
                        default=None)
    parser.add_argument("-f", "--cachefile", help="Cachefile for results " +
                        " [None]", default=None)
    results = parser.parse_args()
    return results


if __name__ == "__main__":
    standalone()



import logging
import os
from jupyter_client.localinterfaces import public_ips
from urllib.parse import urlparse
from .. import Singleton
from ..utils import (str_bool, listify, make_logger, get_execution_namespace,
                     sanitize_dict)


class LSSTConfig(metaclass=Singleton):
    '''LSSTConfig is a Singleton bag of attributes to hold
    important-to-the-LSST JupyterHub environment variables and default
    values.
    '''

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.source = kwargs.pop('source', 'environment')
        if self.source != 'environment':
            raise ValueError(
                "'environment' is currently only allowed config source!")
        # To make it easy to read settings from a document later on...
        if self.source == 'environment':
            self.load_from_environment()
        if self.debug:
            self.log.setLevel(logging.DEBUG)
        self.create_derived_settings()

    def load_from_environment(self):
        '''Populate attributes from environment variables.
        '''
        self.debug = str_bool(os.getenv('DEBUG'))
        # Authentication parameters
        self.authenticator_type = (os.getenv('AUTH_PROVIDER') or
                                   os.getenv('OAUTH_PROVIDER') or
                                   'github')
        self.oauth_client_id = os.getenv('OAUTH_CLIENT_ID')
        self.oauth_client_secret = os.getenv('OAUTH_CLIENT_SECRET')
        self.oauth_callback_url = os.getenv('OAUTH_CALLBACK_URL')
        # Authenticator-specific parameters
        # Yes, really "sign_in".  Blame BVan.
        self.jwt_logout_url = os.getenv("LOGOUT_URL") or '/oauth2/sign_in'
        self.jwt_signing_certificate = (os.getenv('JWT_SIGNING_CERTIFICATE') or
                                        '/opt/jwt/signing-certificate.pem')
        self.cilogon_host = os.getenv('CILOGON_HOST') or 'cilogon.org'
        self.cilogon_skin = os.getenv('CILOGON_SKIN') or 'LSST'
        self.cilogon_idp = os.getenv('CILOGON_IDP_SELECTION')
        self.cilogon_allowlist = listify(os.getenv('CILOGON_GROUP_WHITELIST'))
        self.cilogon_denylist = listify(os.getenv('CILOGON_GROUP_DENYLIST'))
        self.strict_ldap_groups = str_bool(os.getenv('STRICT_LDAP_GROUPS'))
        self.github_host = os.getenv('GITHUB_HOST') or 'github.com'
        self.github_allowlist = listify(
            os.getenv('GITHUB_ORGANIZATION_WHITELIST'))
        self.github_denylist = listify(
            os.getenv('GITHUB_ORGANIZATION_DENYLIST'))
        # Settings for Options Form
        self.form_selector_title = (os.getenv('LAB_SELECTOR_TITLE') or
                                    'Container Image Selector')
        self.form_template = (os.getenv('OPTIONS_FORM_TEMPLATE') or
                              ('/opt/lsst/software/jupyterhub/templates/' +
                               'options_form.template.html'))
        self.form_sizelist = (listify(os.getenv('OPTIONS_FORM_SIZELIST')) or
                              ['tiny', 'small', 'medium', 'large'])
        self.max_scan_delay = int(os.getenv('MAX_SCAN_DELAY', 300))
        self.initial_scan_interval = float(
            os.getenv('INITIAL_SCAN_INTERVAL', 0.2))
        self.max_scan_interval = float(os.getenv('MAX_SCAN_INTERVAL', 5.0))
        self.tiny_cpu = float(os.getenv('TINY_MAX_CPU', 0.5))
        self.mb_per_cpu = int(os.getenv('MB_PER_CPU', 2048))
        self.size_index = int(os.getenv('SIZE_INDEX', 1))
        # Settings for Quota Manager
        self.resource_map = (os.getenv('RESOURCE_MAP') or
                             '/opt/lsst/software/jupyterhub/resources/' +
                             'resourcemap.json')
        self.max_dask_workers = int(os.getenv('MAX_DASK_WORKERS', 25))
        # Settings for Volume Manager
        self.volume_definition_file = (os.getenv('VOLUME_DEFINITION_FILE') or
                                       ('/opt/lsst/software/jupyterhub/' +
                                        'mounts/mountpoints.json'))
        # Hub settings for Lab spawning
        self.lab_default_image = (os.getenv('LAB_IMAGE') or
                                  "lsstsqre/sciplat-lab:latest")
        self.mem_limit = os.getenv('LAB_MEM_LIMIT') or '2048M'
        self.cpu_limit = os.getenv('LAB_CPU_LIMIT') or '1.0'
        self.mem_guarantee = os.getenv('LAB_MEM_GUARANTEE') or '1M'
        self.cpu_guarantee = os.getenv('LAB_CPU_GUARANTEE') or '0.02'
        self.lab_size_range = os.getenv('LAB_SIZE_RANGE') or '4.0'
        self.cull_timeout = os.getenv('LAB_CULL_TIMEOUT') or '64800'
        self.cull_policy = os.getenv('LAB_CULL_POLICY') or 'idle:remote'
        self.allow_dask_spawn = str_bool(os.getenv('ALLOW_DASK_SPAWN'))
        self.restrict_dask_nodes = os.getenv('RESTRICT_DASK_NODES')
        self.restrict_lab_nodes = os.getenv('RESTRICT_LAB_NODES')
        self.lab_nodejs_max_mem = os.getenv('LAB_NODEJS_MAX_MEM') or '6144'
        self.external_hub_url = os.getenv('EXTERNAL_HUB_URL')
        self.hub_route = os.getenv('HUB_ROUTE') or '/'
        self.external_instance_url = os.getenv('EXTERNAL_INSTANCE_URL')
        self.firefly_route = os.getenv('FIREFLY_ROUTE') or '/firefly'
        self.js9_route = os.getenv('JS9_ROUTE') or '/js9'
        self.api_route = os.getenv('API_ROUTE') or '/api'
        self.tap_route = os.getenv('TAP_ROUTE') or '/api/tap'
        self.soda_route = os.getenv('SODA_ROUTE') or '/api/image/soda'
        self.workflow_route = os.getenv('WORKFLOW_ROUTE') or '/workflow'
        self.external_firefly_route = os.getenv('EXTERNAL_FIREFLY_ROUTE')
        self.external_js9_route = os.getenv('EXTERNAL_JS9_ROUTE')
        self.external_api_route = os.getenv('EXTERNAL_API_ROUTE')
        self.external_tap_route = os.getenv('EXTERNAL_TAP_ROUTE')
        self.external_soda_route = os.getenv('EXTERNAL_SODA_ROUTE')
        self.external_workflow_route = os.getenv('EXTERNAL_WORKFLOW_ROUTE')
        self.auto_repo_urls = os.getenv('AUTO_REPO_URLS')
        # Prepuller settings
        self.lab_repo_owner = os.getenv('LAB_REPO_OWNER') or 'lsstsqre'
        self.lab_repo_name = os.getenv('LAB_REPO_NAME') or 'sciplat-lab'
        self.lab_repo_host = os.getenv('LAB_REPO_HOST') or 'hub.docker.com'
        self.prepuller_namespace = (os.getenv('PREPULLER_NAMESPACE') or
                                    get_execution_namespace())
        self.prepuller_experimentals = int(
            os.getenv('PREPULLER_EXPERIMENTALS', 0))
        self.prepuller_dailies = int(os.getenv('PREPULLER_DAILIES', 3))
        self.prepuller_weeklies = int(os.getenv('PREPULLER_WEEKLIES', 2))
        self.prepuller_releases = int(os.getenv('PREPULLER_RELEASES', 1))
        self.prepuller_cachefile = os.getenv('PREPULLER_CACHEFILE',
                                             (os.getenv('HOME') +
                                              '/repo-cache.json'))
        # Fileserver settings
        self.fileserver_host = (os.getenv('EXTERNAL_FILESERVER_IP') or
                                os.getenv('FILESERVER_SERVICE_HOST'))
        # Reaper settings
        self.reaper_user = os.getenv('IMAGE_REAPER_USER')
        self.reaper_password = os.getenv('IMAGE_REAPER_PASSWORD')
        # Hub internal settings
        my_ip = public_ips()[0]
        self.helm_tag = os.getenv('HELM_TAG')
        self.hub_host = os.getenv('HUB_SERVICE_HOST') or my_ip
        self.hub_api_port = os.getenv('HUB_SERVICE_PORT_API') or 8081
        self.proxy_host = os.getenv('PROXY_SERVICE_HOST') or my_ip
        self.proxy_api_port = os.getenv('PROXY_SERVICE_PORT_API') or 8001
        self.session_db_url = os.getenv('SESSION_DB_URL')
        # These have to be set post-initialization to avoid a circular
        #  dependency.
        self.authenticator_class = None
        self.spawner_class = None

    def create_derived_settings(self):
        '''Create further settings from passed-in ones.
        '''
        self.proxy_api_url = 'http://{}:{}'.format(
            self.proxy_host, self.proxy_api_port)
        if self.github_host == 'github.com':
            self.github_api = 'api.github.com'
        else:
            self.github_api = "{}/api/v3".format(self.github_host)
        audience = None
        callback_url = self.oauth_callback_url
        if callback_url:
            netloc = urlparse(callback_url).netloc
            scheme = urlparse(callback_url).scheme
            if netloc and scheme:
                audience = scheme + "://" + netloc
        if not audience:
            audience = self.oauth_client_id or ''
        self.audience = audience
        self.bind_url = 'http://0.0.0.0:8000{}'.format(self.hub_route)
        self.hub_bind_url = 'http://0.0.0.0:8081'.format(self.hub_route)
        self.hub_connect_url = 'http://{}:{}{}'.format(self.hub_host,
                                                       self.hub_api_port,
                                                       self.hub_route)
        mm = self.lab_nodejs_max_mem
        self.lab_node_options = None
        if mm:
            self.lab_node_options = "--max-old-space-size={}".format(mm)
        while self.hub_route.endswith('/') and self.hub_route != '/':
            self.hub_route = self.hub_route[:-1]
        if not self.external_hub_url:
            oauth_callback = self.oauth_callback_url
            endstr = '/hub/oauth_callback'
            if oauth_callback and oauth_callback.endswith(endstr):
                self.external_hub_url = oauth_callback[:-len(endstr)]
        if not self.external_instance_url:
            ehu = self.external_hub_url
            if ehu:
                if ehu.endswith(self.hub_route):
                    self.external_instance_url = ehu[:-len(self.hub_route)]

    def dump(self):
        '''Return dict for pretty printing.
        '''
        myvars = vars(self)
        sanitized = sanitize_dict(myvars, ['oauth_client_secret',
                                           'reaper_password',
                                           'session_db_url'])
        # Stringify classrefs
        for key in ["log", "authenticator_class", "spawner_class"]:
            val = sanitized.get(key)
            if val:
                sanitized[key] = str(val)
        return sanitized



'''This is a mixin class for authenticators for common LSST functionality.
'''
from jupyterhub.auth import Authenticator
from .. import LSSTMiddleManager
from ..config import LSSTConfig


class LSSTAuthenticator(Authenticator):
    '''We create an LSST Manager structure on startup; this is an
    LSSTMiddleManager controlling a set of other managers: auth, env,
    namespace, quota, and volume.

    All LSSTAuthenticator subclasses are expected to create two new fields in
     auth_state:
      * auth_state['group_map'], which contains a dict mapping group name
        (the key) to a group ID number (the value).  GIDs may be strings or
        integers.
      * auth_state['uid'], which contains a string or an integer with the
        user's numeric UID.
    '''
    enable_auth_state = True
    delete_invalid_users = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lsst_mgr = LSSTMiddleManager(parent=self, config=LSSTConfig())

    def resolve_cilogon(self, membership):
        '''Shared between CILogon and JWT (which uses CILogon as its backing
        store) and thus appearing here.  Returns a uid, groupmap pair
        suitable for insertion into auth_state; both uid and group
        values are strings.
        '''
        am = self.lsst_mgr.auth_mgr
        cfg = self.lsst_mgr.config
        groupmap = {}
        for grp in membership['isMemberOf']:
            name = grp['name']
            gid = grp.get('id')
            if not id and not cfg.strict_ldap_groups:
                gid = am.get_fake_gid()
            if gid:
                groupmap[name] = str(gid)
        uid = str(membership['uidNumber'])
        return uid, groupmap

    def dump(self):
        '''Return dict suitable for pretty-printing.
        '''
        ad = {"enable_auth_state": self.enable_auth_state,
              "delete_invalid_users": self.delete_invalid_users,
              "login_handler": str(self.login_handler),
              "lsst_mgr": self.lsst_mgr.dump()
              }
        return ad



'''Classes and tool for scanning Docker repositories.
'''
from .scanrepo import ScanRepo
from .standalone import standalone
from .singletonscanner import SingletonScanner
__all__ = [ScanRepo, SingletonScanner, standalone]



from ..utils import make_logger
from .authmanager import LSSTAuthManager
from .envmanager import LSSTEnvironmentManager
from .namespacemanager import LSSTNamespaceManager
from .optionsformmanager import LSSTOptionsFormManager
from .quotamanager import LSSTQuotaManager
from .volumemanager import LSSTVolumeManager


class LSSTMiddleManager(object):
    '''The LSSTMiddleManager is a class that holds references to various
    LSST-specific management objects and delegates requests to them.
    The idea is that an LSST Spawner, or an LSST Workflow Manager,
    could instantiate a single LSSTMiddleManager, which would then be
    empowered to perform all LSST-specific operations, reducing
    configuration complexity.
    '''
    parent = None
    config = None
    authenticator = None
    spawner = None
    user = None
    api = None
    rbac_api = None

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating LSSTMiddleManager")
        self.parent = kwargs.pop('parent')
        self.log.info(
            "Parent of LSST Middle Manager is '{}'".format(self.parent))
        self.config = kwargs.pop('config')
        self.authenticator = self.parent
        self.auth_mgr = LSSTAuthManager(parent=self)
        self.env_mgr = LSSTEnvironmentManager(parent=self)
        self.namespace_mgr = LSSTNamespaceManager(parent=self)
        self.optionsform_mgr = LSSTOptionsFormManager(parent=self)
        self.quota_mgr = LSSTQuotaManager(parent=self)
        self.volume_mgr = LSSTVolumeManager(parent=self)

    def ensure_resources(self):
        '''Delegate to namespace manager.
        '''
        self.namespace_mgr.ensure_namespace()

    def dump(self):
        '''Return contents dict to pretty-print.
        '''
        md = {"parent": str(self.parent),
              "authenticator": str(self.authenticator),
              "spawner": str(self.spawner),
              "user": str(self.user),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              "config": self.config.dump(),
              "auth_mgr": self.auth_mgr.dump(),
              "env_mgr": self.env_mgr.dump(),
              "optionsform_mgr": self.optionsform_mgr.dump(),
              "quota_mgr": self.quota_mgr.dump(),
              "volume_mgr": self.volume_mgr.dump()
              }
        return md



import json
import os
from kubernetes.client import V1ResourceQuotaSpec
from kubernetes.client.rest import ApiException
from kubernetes import client

from ..utils import make_logger


class LSSTQuotaManager(object):
    '''Quota support for LSST LSP Jupyterlab and Dask pods.
    '''
    quota = {}
    custom_resources = {}
    resourcemap = None

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating LSSTQuotaManager")
        self.parent = kwargs.pop('parent')

    def define_resource_quota_spec(self):
        '''We're going to return a resource quota spec that checks whether we
        have a custom resource map and uses that information.  If we do not
        then we use the quota from our parent's config object.

        Note that you could get a lot fancier, and check the user group
        memberships to determine what class a user belonged to, or some other
        more-sophisticated-than-one-size-fits-all quota mechanism.
        '''
        self.log.debug("Calculating default resource quotas.")
        om = self.parent.optionsform_mgr
        sizemap = om.sizemap
        # sizemap is an ordered dict, and we want the last-inserted one,
        #  which is the biggest
        big = list(sizemap.keys())[-1]
        cpu = sizemap[big]['cpu']
        cfg = self.parent.config
        max_dask_workers = 0
        if cfg.allow_dask_spawn:
            max_dask_workers = cfg.max_dask_workers
        mem_per_cpu = cfg.mb_per_cpu
        total_cpu = (1 + max_dask_workers) * cpu
        total_mem = str(int(total_cpu * mem_per_cpu + 0.5)) + "Mi"
        total_cpu = str(int(total_cpu + 0.5))
        self.log.debug("Default quota sizes: CPU %r, mem %r" % (
            total_cpu, total_mem))
        self._set_custom_user_resources()
        if self.custom_resources:
            self.log.debug("Have custom resources.")
            cpuq = self.custom_resources.get("cpu_quota")
            if cpuq:
                self.log.debug("Overriding CPU quota.")
                total_cpu = str(cpuq)
            memq = self.custom_resources.get("mem_quota")
            if memq:
                self.log.debug("Overriding memory quota.")
                total_mem = str(memq) + "Mi"
        self.log.debug("Determined quota sizes: CPU %r, mem %r" % (
            total_cpu, total_mem))
        qs = V1ResourceQuotaSpec(
            hard={"limits.cpu": total_cpu,
                  "limits.memory": total_mem})
        self.quota = qs.hard

    def _set_custom_user_resources(self):
        '''Create custom resource definitions for user.
        '''
        if not self.resourcemap:
            self.log.debug("No resource map found; generating.")
            self._create_resource_map()
        if not self.resourcemap:
            self.log.warning("No resource map; cannot generate quota.")
            return
        resources = {
            "size_index": 0,
            "cpu_quota": 0,
            "mem_quota": 0
        }
        gnames = self.parent.user.groups
        uname = self.parent.user.escaped_name
        for resdef in self.resourcemap:
            apply = False
            if resdef.get("disabled"):
                continue
            candidate = resdef.get("resources")
            if not candidate:
                continue
            self.log.debug(
                "Considering candidate resource map {}".format(resdef))
            ruser = resdef.get("user")
            rgroup = resdef.get("group")
            if ruser and ruser == uname:
                self.log.debug("User resource map match.")
                apply = True
            if rgroup and rgroup in gnames:
                self.log.debug("Group resource map match.")
                apply = True
            if apply:
                for fld in ["size_index", "cpu_quota", "mem_quota"]:
                    vv = candidate.get(fld)
                    if vv and vv > resources[fld]:
                        resources[fld] = vv
                    self.log.debug(
                        "Setting custom resources '{}'".format(resources) +
                        "for user '{}'".format(uname))
                    self.custom_resources = resources

    def _create_resource_map(self):
        resource_file = self.parent.config.resource_map
        if not os.path.exists(resource_file):
            nf_msg = ("Could not find resource definition file" +
                      " at '{}'".format(resource_file))
            self.log.error(nf_msg)
            return
        with open(resource_file, "r") as rf:
            resmap = json.load(rf)
        self.resourcemap = resmap

    # Brought in from namespacedkubespawner
    def ensure_namespaced_resource_quota(self, quotaspec):
        '''Create K8s quota object if necessary.
        '''
        self.log.debug("Entering ensure_namespaced_resource_quota()")
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.api
        if namespace == "default":
            self.log.error("Will not create quota for default namespace!")
            return
        quota = client.V1ResourceQuota(
            metadata=client.V1ObjectMeta(
                name="quota",
            ),
            spec=quotaspec
        )
        self.log.info("Creating quota: %r" % quota)
        try:
            api.create_namespaced_resource_quota(namespace, quota)
        except ApiException as e:
            if e.status != 409:
                self.log.exception("Create resourcequota '%s'" % quota +
                                   "in namespace '%s' " % namespace +
                                   "failed: %s", str(e))
                raise
            else:
                self.log.debug("Resourcequota '%r' " % quota +
                               "already exists in '%s'." % namespace)

    def destroy_namespaced_resource_quota(self):
        '''Destroys the Kubernetes namespaced resource quota.
        You don't usually have to call this, since it will get
        cleaned up as part of namespace deletion.
        '''
        namespace = self.parent.namespace_mgr.namespace
        api = self.parent.api
        qname = "quota-" + namespace
        dopts = client.V1DeleteOptions()
        self.log.info("Deleting resourcequota '%s'" % qname)
        api.delete_namespaced_resource_quota(
            qname, namespace, dopts)

    def dump(self):
        '''Return contents dict for pretty-printing/aggregation.
        '''
        qd = {"parent": str(self.parent),
              "quota": self.quota,
              "custom_resources": self.custom_resources,
              "resourcemap": self.resourcemap}
        return qd
